require_relative "types"

class Reader
  def initialize(tokens, position = 0)
    @tokens = tokens
    @position = position
  end

  def next
    @position = @position + 1
    @tokens[@position-1]
  end

  def peek
    @tokens[@position]
  end
end

def read_str(str)
  tokens = tokenize str
  reader = Reader.new tokens
  read_form reader
end

def tokenize(str)
  tokens = []
  regex = %r{[\s,]*(~@|[\[\]{}()'`~^@]|"(?:\\.|[^\\"])*"?|;.*|[^\s\[\]{}('"`,;)]*)}
  str.scan(regex) do |token|
    tokens << token[0]
  end
  tokens.pop
  tokens
end

def read_form(reader)
  token = reader.peek
  return case token[0]
  when ";"
    reader.next
    Comment.new
  when "@"
    reader.next
    List.new [:deref, read_form(reader)]
  when "'"
    reader.next
    List.new [:quote, read_form(reader)]
  when "`"
    reader.next
    List.new [:quasiquote, read_form(reader)]
  when "~"
    reader.next
    if token == "~@" then return List.new [:"splice-unquote", read_form(reader)] end
    List.new [:unquote, read_form(reader)]
  when "^"
    reader.next
    meta = read_form(reader)
    List.new [:"with-meta", read_form(reader), meta]
  when "("
    read_list reader
  when "["
    read_vector reader
  when "{"
    read_hashmap reader
  else 
    read_atom reader
  end
end

def read_list(reader)
  list = List.new
  reader.next
  while true
    token = reader.peek
    if not token
      raise "Expected ')', got EOF"
    end
    if token == ")"
      reader.next
      return list
    else
      elem = read_form reader
      if elem.class != Comment then list.push elem end
    end
  end
end

def read_vector(reader)
  vector = Vector.new
  reader.next
  while true
    token = reader.peek
    if not token
      raise "Expected ']', got EOF"
    end
    if token == "]"
      reader.next
      return vector
    else
      vector.push read_form reader
    end
  end
end

def read_hashmap(reader)
  list = []
  reader.next
  while true
    token = reader.peek
    if not token
      raise "Expected '}', got EOF"
    end
    if token == "}"
      reader.next
      return Hash[list.each_slice(2).to_a]
    else
      list.push read_form reader
    end
  end
end

def parse_str(t)
  return t[1..-2].gsub(/\\./, {"\\\\" => "\\", "\\n" => "\n", "\\\"" => '"'})
end

def read_atom(reader)
  token = reader.next
  return case token
  when /^-?[0-9]+$/
    token.to_i
  when /^-?[0-9][0-9.]*$/
    token.to_f
  when /^"(?:\\.|[^\\"])*"$/
    parse_str(token)
  when /^"/ 
    raise "Expected '\"', got EOF"
  when /^:/
    "\u029e" + token[1..-1]
  when "nil"
    nil
  when "true"
    true
  when "false"
    false
  else
    token.to_sym
  end
end
