require "readline"
require_relative "reader"
require_relative "printer"
require_relative "types"
require_relative "env"
require_relative "core"

def quasiquote(ast)
  case ast
  when List
    if ast[0] == :unquote
      return ast[1]
    else
      result = List.new
      ast.reverse_each do |elt|
        if elt.class == List and elt[0] == :"splice-unquote"
          result = List.new [:concat, elt[1], result]
        else
          result = List.new [:cons, quasiquote(elt), result]
        end
      end
      return result
    end
  when Vector
    result = List.new
      ast.reverse_each do |elt|
        if elt.class == List and elt[0] == :"splice-unquote"
          result = List.new [:concat, elt[1], result]
        else
          result = List.new [:cons, quasiquote(elt), result]
        end
      end
      return List.new [:vec, result]
  when Hash, Symbol
    return List.new [:quote, ast]
  else
    return ast
  end
end


def READ(str)
  return read_str str
end

$repl_env = Env.new
NS.each do |key, expr|
  $repl_env.set(key, expr)
end
$repl_env.set(:eval, lambda {|ast| EVAL(ast, $repl_env)})

def eval_ast(ast, env)
  return case ast
  when Symbol
    env.get(ast)
  when List, Vector
    ast.class.new ast.map { |x| EVAL(x, env) }
  when Hash
    ast.map { |k, v| [k, EVAL(v, env)] }.to_h
  else
    ast
  end
end

def EVAL(ast, env)
  while true
    if ast.class != List then return eval_ast(ast, env) end
    if ast.empty? then return ast end
    
    return case ast[0]
    when :def!
      env.set(ast[1], EVAL(ast[2], env))
    when :"let*" 
      let_env = Env.new env
      ast[1].each_slice(2) {|x| let_env.set(x[0], EVAL(x[1], let_env))}
      env=let_env; ast = ast[2]
      next
    when :do
      eval_ast(List.new(ast[1..-2]), env)
      ast = ast[-1]
      next
    when :if
      if EVAL(ast[1], env)
        ast = ast[2]
      else
        ast = ast[3]
      end
      next
    when :quote
      return ast[1]
    when :quasiquoteexpand
      return quasiquote ast[1]
    when :quasiquote
      ast = quasiquote ast[1]
      next
    when :"fn*"
      fn = lambda { |*args|
        fun_env = Env.new(env, ast[1], args)
        return EVAL(ast[2], fun_env)
      }
      return Fun.new(ast[2], ast[1], env, fn)
    else
      result = eval_ast(ast, env)
      f = result[0]; args = result[1..-1]
      if f.class == Fun
        ast = f.ast
        env = Env.new(f.env, f.params, args)
        next
      else
        return f.call *args
      end
    end
  end
end

def PRINT(str)
  pr_str str
end

def rep(str)
  PRINT(EVAL(READ(str), $repl_env))
end

rep "(def! not (fn* (a) (if a false true)))"
rep "(def! load-file (fn* (f) (eval (read-string (str \"(do \" (slurp f) \"\nnil)\")))))"

if ARGV[1..-1]
  $repl_env.set(:"*ARGV*", List.new(ARGV[1..-1]))
else
  $repl_env.set(:"*ARGV*", List.new)
end

if ARGV[0]
  rep "(load-file \"" + ARGV[0] + "\")"
  exit
end

while buf = Readline.readline("user> ", true)
  begin
    puts rep buf
  rescue Exception => e
    puts "Error: #{e}" 
    puts "\t#{e.backtrace.join("\n\t")}"
  end
end

